#!/usr/bin/env python

"""
Translate the AST of a Python program into a more interpretable representation.

Copyright (C) 2007, 2008, 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *
from micropython.rsvp import *
from micropython.trans import Helper
from micropython.code import Assembler
import compiler.ast

# Program visitors.

class Translation(ASTVisitor, Assembler, Helper):

    "A module translator."

    # Attribute access instructions, for use with the appropriate handlers.

    attribute_load_instructions = (
        LoadAddress, LoadAddressContext, LoadAddressContextCond,
        LoadAttr, LoadAttrIndex, LoadAttrIndexContextCond
        )
    # attribute_store_instructions are defined by the optimiser

    # Name access instructions, for use with the appropriate handlers.

    name_load_instructions = (LoadName, LoadAddress, None)
    name_store_instructions = (StoreName, StoreAddress, StoreAddressContext)

    def __init__(self, module, program):

        """
        Initialise the translation with an inspected 'module' and the 'program'
        container.
        """

        ASTVisitor.__init__(self)
        Assembler.__init__(self, program.optimisations)
        self.visitor = self
        self.module = module

        # Global program dependencies.

        self.program = program
        self.objtable = self.program.get_object_table()
        self.paramtable = self.program.get_parameter_table()
        self.importer = self.program.get_importer()
        self.builtins = self.importer.modules.get("__builtins__")

        # Status flags.

        self.in_exception_handler = 0
        self.in_assignment = 0          # for slicing and subscript

        # Reset the assembler.

        self.reset()

    def __repr__(self):
        return "Translation(%r)" % self.module

    def get_module_code(self):

        """
        Return the top-level module code.
        """

        self.unit = self.module
        self.reset()
        self.add_exception_unit()

        block = self.new_block()
        self.set_block(block)

        # Handle exceptions for the program.

        if self.module.name == "__main__":
            handler_block = self.new_block()
            self.new_op(PushHandler(handler_block))

        # Generate code for the module.

        if self.module.astnode is not None:
            self.dispatch(self.module.astnode)

        # Finish off the translated program if appropriate.

        if self.module.name == "__main__":
            self.set_block(handler_block)
            self.new_op(PopHandler(1))
            self.new_op(Return())

        self.drop_exception_unit()
        self.unit.temp_usage = self.max_temp_position + 1
        self.unit.blocks = self.blocks
        return self.blocks

    def get_code(self, unit):

        "Return the code for the given 'unit'."

        self.unit = unit
        self.reset()
        self.add_exception_unit()

        block = self.new_block()
        self.set_block(block)

        if unit.astnode is not None:
            self.dispatch(unit.astnode)

        self.drop_exception_unit()
        self.unit.temp_usage = self.max_temp_position + 2 # include space for instantiators to expand backwards
        self.unit.blocks = self.blocks
        return self.blocks

    def get_instantiator_code(self, cls):

        "Return the code for the given class 'cls'."

        # Obtain the function object to be populated.

        self.unit = cls.get_instantiator()
        self.reset()

        block = self.new_block()
        self.set_block(block)

        init_method = cls.get_init_method()

        # Make an object and store it in the unused first slot.

        self.make_instance(cls, len(cls.instance_attributes()))
        self.new_op(StoreTemp(0))

        # Invoke the appropriate initialiser.

        self.new_op(LoadFunction(init_method))
        self.new_op(LoadCallable())
        self.new_op(JumpInFrame())

        # Store the object as the result.

        self.new_op(LoadTemp(0)) # load the context from the locals
        self.new_op(Return())

        self.unit.blocks = self.blocks
        return self.blocks

    # Visitor methods.

    def default(self, node, *args):
        raise TranslateError("Node class %r is not supported." % node.__class__)

    # Concrete visitor methods.

    # Binary operators.

    visitAdd        = Helper._visitBinary
    visitBitand     = Helper._visitBinaryBit
    visitBitor      = Helper._visitBinaryBit
    visitBitxor     = Helper._visitBinaryBit
    visitDiv        = Helper._visitBinary
    visitFloorDiv   = Helper._visitBinary
    visitLeftShift  = Helper._visitBinary
    visitMod        = Helper._visitBinary
    visitMul        = Helper._visitBinary
    visitPower      = Helper._visitBinary
    visitRightShift = Helper._visitBinary
    visitSub        = Helper._visitBinary

    # Unary operators.

    visitInvert     = Helper._visitUnary
    visitUnaryAdd   = Helper._visitUnary
    visitUnarySub   = Helper._visitUnary

    # Logical operators.

    def visitAnd(self, node):
        end_block = self.new_block()
        temp_pos = self.reserve_temp()

        for n in node.nodes[:-1]:
            self.dispatch(n)
            self.new_op(StoreTemp(temp_pos))

            self._generateTestBoolean(n, LoadTemp(temp_pos))
            self.new_op(JumpIfFalse(end_block, working="status"))

        self.dispatch(node.nodes[-1])
        self.new_op(StoreTemp(temp_pos))

        self.set_block(end_block)

        # Make a separate instruction to prevent previous temp accesses from
        # being altered by assign_value.

        temp = LoadTemp(temp_pos)
        self.new_op(temp)
        self.discard_temp(temp)

    def visitNot(self, node):
        self.dispatch(node.expr)

        temp = self.optimiser.optimise_temp_storage()
        self.new_op(temp)
        self._generateTestBoolean(node.expr, temp)
        self.discard_temp(temp)

        self.new_op(InvertBoolean(source="status", target="status"))
        self._generateLoadBoolean(node)

    def visitOr(self, node):
        end_block = self.new_block()
        temp_pos = self.reserve_temp()

        for n in node.nodes[:-1]:
            self.dispatch(n)
            self.new_op(StoreTemp(temp_pos))

            self._generateTestBoolean(n, LoadTemp(temp_pos))
            self.new_op(JumpIfTrue(end_block, working="status"))

        self.dispatch(node.nodes[-1])
        self.new_op(StoreTemp(temp_pos))

        self.set_block(end_block)

        # Make a separate instruction to prevent previous temp accesses from
        # being altered by assign_value.

        temp = LoadTemp(temp_pos)
        self.new_op(temp)
        self.discard_temp(temp)

    # Comparisons.

    def visitCompare(self, node):

        """
        _t1 = node.expr
        _t1 op1 _t2 and _t2 op2 _t3 and ...
        """

        end_block = self.new_block()

        self.dispatch(node.expr)
        temp2 = self.optimiser.optimise_temp_storage()

        # NOTE: Replicated by some code in micropython.inspect.visitCompare.

        last_op = node.ops[-1]

        for op in node.ops:
            op_name, next_node = op
            operator_fn = operator_functions.get(op_name)

            # Propagate the arguments as we traverse the construct.

            temp1 = temp2
            self.dispatch(next_node)
            temp2 = self.optimiser.optimise_temp_storage()

            # Use the appropriate mechanism, setting the boolean status for the
            # comparison.

            if operator_fn is not None:

                # Generate function call using evaluated argument and next node.

                temp_fn = self._generateOperatorFunction(op_name)
                self._generateInvocation(temp_fn, (temp1, temp2))
                self.discard_temp(temp_fn)

                temp_result = self.optimiser.optimise_temp_storage()
                self._generateTestBoolean(node, temp_result)
                self.discard_temp(temp_result)

            else:
                # Deal with the special operators.

                if op_name.startswith("is"):
                    self.new_op(temp1)
                    self.record_value()
                    self.new_op(temp2)
                    self.new_op(TestIdentity(target="status"))
                    self.assign_value()
                    self.discard_value()

                elif op_name.endswith("in"):
                    self.new_op(temp2)

                    # Get method on temp2.

                    self._generateAttr(node, "__contains__", self.attribute_load_instructions)
                    temp_method = self.optimiser.optimise_temp_storage()

                    # Add arguments.
                    # NOTE: No support for defaults.

                    self._generateInvocation(temp_method, (temp2, temp1))

                    temp_result = self.get_temp()
                    self._generateTestBoolean(node, temp_result)
                    self.discard_temp(temp_result)

                if op_name.find("not") != -1:
                    self.new_op(InvertBoolean(source="status", target="status"))

            # Test the result and jump to the end block if false.

            if op is not last_op:
                self.new_op(JumpIfFalse(end_block, working="status"))

            # Compilation duties...

            self.discard_temp(temp1)

        self.discard_temp(temp2)

        # With the status set above, produce a boolean result.

        self.set_block(end_block)

        # Yield the appropriate value.

        self._generateLoadBoolean(node)

    # Expressions.

    def visitBackquote(self, node): raise TranslationNotImplementedError("Backquote")

    def visitCallFunc(self, node):

        """
        Evaluate positional arguments, evaluate and store keyword arguments in
        the correct location, then invoke the function.
        """

        # Mark the frame, evaluate the target, generate the call.

        self._startCallFunc()
        self.dispatch(node.node)
        temp_target, target, temp_context = self._generateCallFunc(node.args, node)
        self._doCallFunc(temp_target, target)
        self._endCallFunc(temp_target, temp_context)

    def visitConst(self, node):
        const = self.importer.get_constant(node.value)
        self.new_op(LoadConst(const))

    def visitDict(self, node): raise TranslationNotImplementedError("Dict")

    def visitEllipsis(self, node): raise TranslationNotImplementedError("Ellipsis")

    def visitExec(self, node): raise TranslationNotImplementedError("Exec")

    def visitExpression(self, node): raise TranslationNotImplementedError("Expression")

    def visitGenExpr(self, node): raise TranslationNotImplementedError("GenExpr")

    def visitGenExprFor(self, node): raise TranslationNotImplementedError("GenExprFor")

    def visitGenExprIf(self, node): raise TranslationNotImplementedError("GenExprIf")

    def visitGenExprInner(self, node): raise TranslationNotImplementedError("GenExprInner")

    def visitGetattr(self, node):
        self._visitAttr(node, self.attribute_load_instructions)

    def visitList(self, node):
        self._generateList(node)

    def visitListComp(self, node): raise TranslationNotImplementedError("ListComp")

    def visitListCompFor(self, node): raise TranslationNotImplementedError("ListCompFor")

    def visitListCompIf(self, node): raise TranslationNotImplementedError("ListCompIf")

    def visitName(self, node):
        self._visitName(node, self.name_load_instructions)

    def visitSlice(self, node):
        args = [node.expr]

        if node.lower is None:
            args.append(compiler.ast.Name("None"))
            if node.upper is None:
                args.append(compiler.ast.Name("None"))
            else:
                args.append(node.upper)
        else:
            args.append(node.lower)
            if node.upper is None:
                args.append(compiler.ast.Name("None"))
            else:
                args.append(node.upper)

        temp_fn = self._getOperatorFunction(node, self.in_assignment and "AssSlice" or "Slice")
        self._visitCall(node, temp_fn, args)
        self.discard_temp(temp_fn)

    def visitSubscript(self, node):
        temp_fn = self._getOperatorFunction(node, self.in_assignment and "AssSubscript" or "Subscript")
        self._visitCall(node, temp_fn, [node.expr] + node.subs)
        self.discard_temp(temp_fn)

    def visitTuple(self, node):
        self._generateTuple(node)

    # Definitions.

    def visitAssign(self, node):

        """
        Evaluate the expression from the given 'node' and assign it to the
        associated recipients.
        """

        self.dispatch(node.expr)

        # Record the value and then dispatch to the assignment targets.

        self.record_value(self.has_immediate_usage(node.nodes))

        self.in_assignment = 1

        for n in node.nodes:
            self.dispatch(n)

        self.in_assignment = 0
        self.discard_value()

    def visitAssAttr(self, node):

        "Assign the assignment expression to the recipient 'node'."

        self._visitAttr(node, self.optimiser.get_attribute_store_instructions())
        self.assign_value()

    def visitAssList(self, node):

        """
        Assign items from the assignment expression to each of the recipients
        found within the given 'node'.
        """

        self.new_op(self.expr_temp[-1])
        self._generateAttr(node, "__getitem__", self.attribute_load_instructions)
        temp_getitem = self.optimiser.optimise_temp_storage()

        for i, n in enumerate(node.nodes):
            self._startCallFunc()
            self.new_op(temp_getitem)
            temp_target, target, temp_context = self._generateCallFunc([compiler.ast.Const(i)], node)
            self._doCallFunc(temp_target, target)
            self._endCallFunc()

            # Provide a different source value.
            # NOTE: Permitting immediate usage given that neither name nor
            # NOTE: attribute accesses should involve a function call
            # NOTE: overwriting the above result.

            self.record_value(self.is_immediate_user(n))
            self.dispatch(n)
            self.discard_value()

        self.discard_temp(temp_getitem)

    def visitAssName(self, node):

        "Assign the assignment expression to the recipient 'node'."

        if hasattr(node, "flags") and node.flags == "OP_DELETE":
            raise TranslationNotImplementedError("AssName(OP_DELETE)")

        self._visitName(node, self.name_store_instructions)
        self.assign_value()

        # Add any attribute usage guards.

        self._generateGuards(node)

    visitAssTuple = visitAssList

    def visitAugAssign(self, node):

        # Find the augmented assignment function and attempt to use it.

        temp_fn = self._getOperatorAugAssignFunction(node)
        self._visitCall(node, temp_fn, (node.node, node.expr))
        self.discard_temp(temp_fn)

        # Assign the result to the name.

        self.record_value(1)

        if isinstance(node.node, compiler.ast.Name):
            self.visitAssName(node.node)
        elif isinstance(node.node, compiler.ast.Getattr):
            self.visitAssAttr(node.node)
        else:
            raise TranslationNotImplementedError("AugAssign(Slice or Subscript)")

        self.discard_value()

    def visitClass(self, node):
        if not used_by_unit(node):
            return

        # Store the name.

        self.new_op(LoadClass(node.unit))
        self.record_value()
        self._visitName(node, self.name_store_instructions)
        self.assign_value()
        self.discard_value()

        # Visit the code.

        unit = self.unit
        self.unit = node.unit
        self.dispatch(node.code)
        self.unit = unit

    def visitDecorators(self, node): raise TranslationNotImplementedError("Decorators")

    def visitFrom(self, node): pass

    def visitFunction(self, node):
        if not used_by_unit(node):
            return

        # Only store the name when visiting this node from outside.

        if self.unit is not node.unit:
            self._visitFunctionDeclaration(node)

            # Record the declared function.

            self.record_value()

            self._visitName(node, self.name_store_instructions) # AssName equivalent
            self.assign_value()
            self.discard_value()

        # Visiting of the code occurs when get_code is invoked on this node.

        else:
            self._visitFunctionDefinition(node)

    def visitGlobal(self, node): pass

    def visitImport(self, node): pass

    def visitKeyword(self, node): pass

    def visitLambda(self, node):

        """
        Lambda functions can be represented as globally defined functions
        provided they do not define any default parameter values, since these
        may defined in a non-global scope.

        Where defaults are defined, an object must be created and its content
        defined: the callable member of the object's structure must be set to
        the lambda function definition; each default must be attached to the
        object as an attribute, as is the case with normal functions and
        methods.
        """

        # Produce the reference to this function when visiting this node from
        # outside.

        if self.unit is not node.unit:

            # Provide the declared function.

            self._visitFunctionDeclaration(node)

        # Visiting of the code occurs when get_code is invoked on this node.

        else:
            self._visitFunctionDefinition(node)

    def visitModule(self, node):
        extend = ExtendFrame()
        self.new_op(extend)
        self.dispatch(node.node)
        self.set_frame_usage(node, extend)

    # Statements.

    def visitStmt(self, node):

        "Process the collection of statements provided by 'node'."

        for n in node.nodes:

            # Process the statement.

            self.dispatch(n)

            # Discard temporary storage.

            if self.temp_positions:
                #print "Had temp", self.temp_positions
                self.temp_positions = set()

            # Prevent incorrect optimisation by resetting the optimiser after
            # each statement.

            self.optimiser.reset()

    def visitAssert(self, node): raise TranslationNotImplementedError("Assert")

    def visitBreak(self, node):
        next_block, exit_block = self.get_loop_blocks()
        self.new_op(Jump(exit_block))

    def visitContinue(self, node):
        next_block, exit_block = self.get_loop_blocks()
        self.new_op(Jump(next_block))

    def visitDiscard(self, node):
        self.dispatch(node.expr)
        self.optimiser.optimise_unused_results()

    def visitFor(self, node):
        next_handler_block = self.new_block()
        end_handler_block = self.new_block()
        exit_block = self.new_block()
        next_block = self.new_block()
        else_block = self.new_block()

        # Get the "list" to be iterated over, obtain its iterator.

        self._startCallFunc()
        self.dispatch(node.list)
        self._generateAttr(node, "__iter__", self.attribute_load_instructions)
        temp_target, target, temp_context = self._generateCallFunc([], node)
        self._doCallFunc(temp_target, target)
        self._endCallFunc(temp_target, temp_context)

        # Use a long-lasting temporary storage slot, since any result from the
        # __iter__ method will not remain around for long.

        temp_iterator = self.get_temp()

        # In the loop...

        self.set_block(next_block)

        # Handle exceptions when calling "next"...

        self.add_exception_blocks(next_handler_block, end_handler_block)
        self.new_op(PushHandler(next_handler_block))

        # Use the iterator to get the next value.

        self._startCallFunc()
        self.new_op(temp_iterator)
        self._generateAttr(node, "next", self.attribute_load_instructions)
        temp_target, target, temp_context = self._generateCallFunc([], node)
        self._doCallFunc(temp_target, target)
        self._endCallFunc(temp_target, temp_context)

        # Record the value to be assigned.

        self.record_value()

        # Skip the handler where the call was successful.

        self.new_op(PopHandler(1))
        self.new_op(Jump(end_handler_block))

        # Enter the exception handler.

        self.set_block(next_handler_block)
        self.new_op(PopHandler(1))

        # Disable the handlers.

        self.drop_exception_blocks()

        # Test for StopIteration.

        self.load_builtin("StopIteration", node)
        self.new_op(CheckException(target="status"))
        if node.else_ is not None:
            self.new_op(JumpIfTrue(else_block, working="status"))
        else:
            self.new_op(JumpIfTrue(exit_block, working="status"))

        # Re-raise the exception otherwise.

        self.new_op(RaiseException())

        # After the handler, clear the exception.

        self.set_block(end_handler_block)

        # Assign to the target.

        self.dispatch(node.assign)
        self.discard_value()

        # Process the body with the current next and exit points.

        self.add_loop_blocks(next_block, exit_block)
        self.dispatch(node.body)
        self.drop_loop_blocks()

        # Repeat the loop.

        self.new_op(Jump(next_block))

        # Produce the "else" section.

        if node.else_ is not None:
            self.set_block(else_block)
            self.new_op(ClearException(target="exception"))
            self.dispatch(node.else_)

            # After the loop...

            self.set_block(exit_block)

        else:
            # After the loop...

            self.set_block(exit_block)
            self.new_op(ClearException(target="exception"))

        # Compilation duties...

        self.discard_temp(temp_iterator)

    def visitIf(self, node):
        first = 1
        next_block = None
        exit_block = self.new_block()

        clauses = node.tests + [(None, node.else_)]

        for clause in clauses:
            test, body = clause
            if body is None:
                break

            if not first:
                self.new_op(Jump(exit_block)) # finish last body
                self.set_block(next_block)    # start next test
                next_block = None

            if test is not None:
                self.dispatch(test)

                temp = self.optimiser.optimise_temp_storage()
                self.new_op(temp)
                self._generateTestBoolean(node, temp)

                next_block = self.new_block()
                self.new_op(JumpIfFalse(next_block, working="status"))

            self.dispatch(body)
            first = 0

        if next_block is not None:
            self.set_block(next_block)

        self.set_block(exit_block)

    def visitIfExp(self, node): raise TranslationNotImplementedError("IfExp")

    def visitPass(self, node): pass

    def visitPrint(self, node):
        self._visitPrint(node, "_print")

    def visitPrintnl(self, node):
        self._visitPrint(node, "_printnl")

    def visitRaise(self, node):

        if node.expr1 is not None:

            # NOTE: expr1 only => instance provided

            self.dispatch(node.expr1)

            if node.expr2 is not None:
                temp = self.optimiser.optimise_temp_storage()

                self.dispatch(node.expr2)
                temp_arg = self.optimiser.optimise_temp_storage()

                self._generateInvocation(temp, (temp_arg,))

                self.discard_temp(temp_arg)

            self.new_op(Transfer(source="working", target="exception"))

        self.new_op(RaiseException())

    def visitReturn(self, node):
        if node.value is not None:
            self.dispatch(node.value)
        else:
            self.dispatch(compiler.ast.Name("None"))

        if self.in_exception_handler:
            self.new_op(ClearException(target="exception"))

        # NOTE: Support finally blocks here.

        if self.exception_blocks[-1]:
            self.new_op(PopHandler(len(self.exception_blocks[-1])))

        self.new_op(Return())

    def visitTryExcept(self, node):
        exit_block = self.new_block()
        else_block = self.new_block()
        handler_block = self.new_block()

        self.add_exception_blocks(handler_block, exit_block)

        # Try...
        # Produce the code, then jump to the exit.

        self.new_op(PushHandler(handler_block))
        self.dispatch(node.body)
        self.new_op(PopHandler(1))

        if node.else_ is not None:
            self.new_op(Jump(else_block))
        else:
            self.new_op(Jump(exit_block))

        # Start of handlers.

        self.set_block(handler_block)
        self.new_op(PopHandler(1))

        # Disable the handlers.

        self.drop_exception_blocks()

        for name, assignment, handler in node.handlers:
            next_block = self.new_block()

            # Test the given exception against the current exception.

            if name is not None:
                self.dispatch(name)

                self.new_op(CheckException(target="status"))
                self.new_op(JumpIfFalse(next_block, working="status"))

                # Handle assignment to exception variable.

                if assignment is not None:
                    self.new_op(Transfer(source="working", target="exception"))

                    # Record the value to be assigned.

                    self.record_value()
                    self.dispatch(assignment)
                    self.discard_value()

            # Produce the handler code, then jump to the exit.

            self.in_exception_handler = 1
            self.dispatch(handler)
            self.in_exception_handler = 0

            self.new_op(Jump(exit_block))

            self.set_block(next_block)

        # Unhandled exceptions.

        self.new_op(RaiseException())

        # Optional else clause.

        if node.else_ is not None:
            self.set_block(else_block)
            self.dispatch(node.else_)

        # Clear the exception.

        self.set_block(exit_block)
        self.new_op(ClearException(target="exception"))

    def visitTryFinally(self, node):

        """
        Add finally handler, potentially as an exception handler.
        Generate body, potentially changing return statements so that they do
        not return immediately.
        Generate handler, removing the handler from the active handler list,
        adding instructions which raise active exceptions.
        """

        raise TranslationNotImplementedError("TryFinally")

    def visitWhile(self, node):
        exit_block = self.new_block()
        next_block = self.new_block()
        else_block = self.new_block()

        self.set_block(next_block)
        self.dispatch(node.test)

        temp = self.optimiser.optimise_temp_storage()
        self.new_op(temp)
        self._generateTestBoolean(node, temp)

        if node.else_ is not None:
            self.new_op(JumpIfFalse(else_block, working="status"))
        else:
            self.new_op(JumpIfFalse(exit_block, working="status"))

        self.add_loop_blocks(next_block, exit_block)

        self.dispatch(node.body)
        self.new_op(Jump(next_block))

        if node.else_ is not None:
            self.set_block(else_block)

            self.dispatch(node.else_)

        self.set_block(exit_block)

        self.drop_loop_blocks()

    def visitWith(self, node): raise TranslationNotImplementedError("With")

    def visitYield(self, node): raise TranslationNotImplementedError("Yield")

# vim: tabstop=4 expandtab shiftwidth=4
